/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j;

import static com.google.common.collect.Sets.newHashSet;
import static org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.OperatorUtil.retrieveCopyModule;
import static org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.OperatorUtil.retrieveOperatorModuleType;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.af3.exploration.alg.util.ExplorationFeatureUtils.getEnabledApplicableFeatures;
import static org.fortiss.af3.exploration.alg.util.ExplorationFeatureUtils.getEnabledEvaluators;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.exploration.alg.dse.backend.ExplorerBackendBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.GuiceDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.mating.MatingCrossoverMutateRepairModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.DseProcessProblemModule;
import org.fortiss.af3.exploration.alg.dse.module.DecoderModule;
import org.fortiss.af3.exploration.alg.dse.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SolutionState;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.opt4j.core.Genotype;
import org.opt4j.core.Individual;
import org.opt4j.core.Objective;
import org.opt4j.core.Value;
import org.opt4j.core.common.completer.IndividualCompleterModule;
import org.opt4j.core.config.Task;
import org.opt4j.core.config.TaskStateListener;
import org.opt4j.core.optimizer.Archive;
import org.opt4j.core.optimizer.Control;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.problem.Evaluator;
import org.opt4j.operators.crossover.CrossoverModule;
import org.opt4j.operators.mutate.MutateModule;
import org.opt4j.optimizers.ea.EvolutionaryAlgorithmModule;
import org.opt4j.optimizers.ea.Spea2Module;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.inject.CreationException;
import com.google.inject.Injector;
import com.google.inject.Module;

/**
 * Opt4J optimization backend that defines the DSE task for this particular optimization framework.
 * Here, the problem definition, the variation modules, the optimizer, and the solution format are
 * defined. Additionally, the optimization task is started here and the results are returned.
 * 
 * @author huang
 */
public class Opt4JExplorerBackend extends ExplorerBackendBase<EAOptimizer> {

	/** Opt4J optimization process */
	private Opt4JDseTask opt4JTask;

	/** EA parameters. */
	private EAOptimizer opt4JParameters;

	/**
	 * Contains the problem module that includes the evaluators associated with the goals of the
	 * DSE.
	 */
	private DseProcessProblemModule archExplorationProblemModule;

	/**
	 * Constructor that initializes the Opt4J backend, and sets optimizer independent parameters
	 * and the goal specification
	 * 
	 * @throws Exception
	 */
	public Opt4JExplorerBackend(DseSpecification dse, ExplorationSpecification expSpec,
			Map<ExplorationTarget<?>, ExplorationTarget<?>> targetAssocMap, Injector dseInjector,
			Set<Class<? extends Phenotype>> requestedSolutions, IProgressMonitor progressMonitor)
			throws Exception {
		// Optimizer independent configuration
		super(dse, expSpec, targetAssocMap, dseInjector, requestedSolutions, progressMonitor);

		if(dse.getOptimizer() instanceof EAOptimizer) {
			init((EAOptimizer)dse.getOptimizer());
		} else {
			throw new Exception(
					"Opt4JExplorerBackend cannot be configured with optimizer dependent configuration of type " +
							dse.getClass());
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void init(EAOptimizer opt4JParameters) throws ExplorationException {
		this.opt4JParameters = opt4JParameters;

		// checkDseProcessExploration(requestedSolutions);

		System.out.println("Initializing EA optimizer (generations: " + "" +
				opt4JParameters.getIterations() + ", alpha: " + opt4JParameters.getAlpha() + ").");

		// Instantiate modules that describe the optimization problem
		Collection<Module> modules = new ArrayList<Module>();
		modules.add(this);

		// Evolutionary algorithm module
		EvolutionaryAlgorithmModule evolutionaryAlgorithmModule = new EvolutionaryAlgorithmModule();
		evolutionaryAlgorithmModule.setGenerations(opt4JParameters.getIterations());
		evolutionaryAlgorithmModule.setAlpha(opt4JParameters.getAlpha());
		evolutionaryAlgorithmModule.setMu(opt4JParameters.getMu());
		evolutionaryAlgorithmModule.setCrossoverRate(0.3);
		modules.add(evolutionaryAlgorithmModule);

		// SPEA2 specific module
		Spea2Module spea2Module = new Spea2Module();
		spea2Module.setTournament(opt4JParameters.getSpea2Tournament());
		modules.add(spea2Module);
		// Nsga2Module nsga2Module = new Nsga2Module();
		// modules.add(nsga2Module);

		MatingCrossoverMutateRepairModule mating = new MatingCrossoverMutateRepairModule();
		modules.add(mating);

		// TODO: Replace {@link MultiEvaluator} with {@link ExplorationGoalPriorityMultiEvaluator}
		// that is
		// not based on a @Priority annotation, but on the priority attribute provided by the goal
		// specifications for the different optimization dimensions.
		// AbstractModule multiEvaluatorModule = new AbstractModule() {
		// @Override
		// protected void configure() {
		// bind(MultiEvaluator.class).to(ExplorationGoalPriorityMultiEvaluator.class);
		//
		// }
		// };
		// modules.add(multiEvaluatorModule);

		archExplorationProblemModule =
				new DseProcessProblemModule(dse, dseInjector, requestedSolutions);
		modules.add(archExplorationProblemModule);

		Collection<IExplorationFeature> enabledFeatures =
				getEnabledApplicableFeatures(dse, dseInjector, true);
		for(IExplorationFeature feature : enabledFeatures) {
			for(Class<? extends Genotype> genotype : feature.getGenotypes()) {
				modules.add(retrieveOperatorModuleType(MutateModule.class, genotype, feature));
				modules.add(retrieveOperatorModuleType(CrossoverModule.class, genotype, feature));
				modules.add(retrieveCopyModule(genotype, feature));
			}
		}

		// Define a completer module that "controls" the execution of the decode and the evaluation
		// phases. It also allows to parallelize the execution of these phases.
		IndividualCompleterModule comp = new IndividualCompleterModule();

		if(opt4JParameters.isParallelExecution()) {
			comp.setType(IndividualCompleterModule.Type.PARALLEL);
			comp.setThreads(Runtime.getRuntime().availableProcessors());
			// comp.setThreads(1);
		}
		modules.add(comp);

		// Create and initialize optimization task
		opt4JTask = new Opt4JDseTask(false);
		opt4JTask.init(dseInjector);
		opt4JTask.init(modules);
		dseInjector.getInstance(Opt4JDseTaskProvider.class).setInjector(opt4JTask);

		if(progressMonitor != null) {
			opt4JTask.addStateListener(new TaskStateListener() {

				@Override
				public void stateChanged(Task task) {
					// stateChanged() is invoked after every generation
					progressMonitor.worked(1);

					if(progressMonitor.isCanceled()) {
						System.out.println("DSE has been canceled.");
						Control control = opt4JTask.getInstance(Control.class);
						control.doStop();
					}
				}
			});
		}
	}

	/**
	 * Consistency check for the DSE process configuration.
	 * An {@link Injector} for the decoding phase is setup to check whether the bindings of the
	 * {@link DecoderModule}s are correct given the current configuration of
	 * {@link IExplorationFeature}s and the desired result types.
	 * <p>
	 * Typical errors that are detected include:
	 * <ul>
	 * <li>Circular dependencies</li>
	 * <li>Two or more {@link DecoderModule}s that provide the same {@link Phenotype}</li>
	 * </ul>
	 */
	@SuppressWarnings("unchecked")
	private void checkDseProcessExploration(Set<Class<? extends Phenotype>> resultTypes) {
		Collection<IExplorationFeature> enabledFeatures =
				getService(IExplorationFeatureService.class).getExplorationFeatures(
						dse.getFeatures());
		Collection<Class<? extends DecoderModule<?>>> enabledDecoders = newHashSet();
		enabledFeatures.forEach(f -> f.getDecoders().forEach(
				dec -> enabledDecoders.add((Class<? extends DecoderModule<?>>)dec.getClass())));
		// Note: The interceptor is not used: It is only activated if a Phenotype instance is
		// requested from the decoders.
		GuiceDecoder guiceDecoder = dseInjector.getInstance(GuiceDecoder.class);
		guiceDecoder.setEnabledDecoders(enabledDecoders);
		buildCheckedDecodingInjector(guiceDecoder);
	}

	/**
	 * Creates an decoding {@link Injector} for the DSE process decoding that throws an
	 * {@link ExplorationRuntimeException} if the {@link DecoderModule} setup is invalid.
	 * 
	 * @param guiceDecoder
	 *            {@link Decoder} installing its sub-decoders.
	 * @return the {@link Injector} of the DSE process top-most {@link Decoder}.
	 */
	private Injector buildCheckedDecodingInjector(GuiceDecoder guiceDecoder) {
		Injector decodingInjector = null;
		try {
			decodingInjector = dseInjector.createChildInjector(guiceDecoder);
		} catch(CreationException e) {
			throw new ExplorationRuntimeException(
					"Error while configuring the DSE process decoder chain.\n"
							+ "See the subsequent error message for details. Typical configuration faults include:\n"
							+ "1) 2 or more decoders attemp to provide the same type.\n" + "2) ...",
					e);
		}
		return decodingInjector;
	}

	/** {@inheritDoc} */
	@Override
	public ExplorationSolution explore() throws Exception {

		System.out.println("Running EA optimization...");
		if(progressMonitor != null) {
			progressMonitor.beginTask("EA optimization", opt4JParameters.getIterations());
		}

		ExplorationSolution rval = null;
		Exception ex = null;
		try {
			// Execute optimization
			opt4JTask.execute();

			Archive archive = opt4JTask.getInstance(Archive.class);
			BiMap<ExplorationTarget<?>, Objective> targetObjectiveAssoc =
					createObjectiveTargetAssoc(getEnabledEvaluators(expSpec));

			// Only add feasible solutions to the returned solutions, i.e. solutions that do not
			// violate constraints.
			filterValidSolutions(archive, targetObjectiveAssoc);

			// NOTE: Implicitly generated ExplorationTargets are removed here as the solution is
			// constructed by iterating over the original input ExplorationSpecification. The
			// removal of ExplorationTargets does not affect the found Pareto-Front.
			Opt4JExplorationSolutionSet opt4JSolution =
					new Opt4JExplorationSolutionSet(archive, copyToOrigTargetMap,
							targetObjectiveAssoc);
			rval = opt4JSolution.getExplorationSolution();
			rval.setSolutionState(SolutionState.OPTIMIZED);
		} catch(Exception e) {
			ex = e;
		} finally {
			opt4JTask.close();
		}

		if(ex != null) {
			throw ex;
		}

		return rval;
	}

	/**
	 * Prunes the given archive such that it contains only feasible solutions, i.e.
	 * {@link Individual} that do not violate any constraint.
	 */
	protected void filterValidSolutions(Archive archive,
			Map<ExplorationTarget<?>, Objective> targetObjectiveAssoc) {
		// Check each individual for constraint violations and only further process those which
		// violate none.
		Collection<Individual> infeasibleSolutions = new ArrayList<>();
		for(Individual curIndividual : archive) {
			for(ExplorationConstraint<?> target : pickInstanceOf(ExplorationConstraint.class,
					expSpec.getTargets())) {
				Objective objective = targetObjectiveAssoc.get(target);
				Value<?> value = curIndividual.getObjectives().get(objective);

				if(!validateSolution(target, value)) {
					infeasibleSolutions.add(curIndividual);
				}
			}
		}

		archive.removeAll(infeasibleSolutions);
	}

	/** Creates a map associating {@link ExplorationTarget}s with their Opt4J equivalents. */
	private <RT, T extends ExplorationTarget<RT>, S extends Phenotype, E extends Evaluator<?>>
			BiMap<ExplorationTarget<?>, Objective> createObjectiveTargetAssoc(
					Set<IEvaluatorModule<?, ?, ?>> evaluatorModules) {
		BiMap<ExplorationTarget<?>, Objective> targetObjectiveAssoc =
				HashBiMap.create(evaluatorModules.size());
		for(IEvaluatorModule<?, ?, ?> module : evaluatorModules) {
			Objective evalObjective = module.getObjective();
			try {
				targetObjectiveAssoc.put(module.getExplorationTarget(), evalObjective);
			} catch(Exception e) {
				e.printStackTrace();
			}
		}
		return targetObjectiveAssoc;
	}
}
